#!/usr/bin/perl -w
#
# readinfo - Read fields from a formatted text file.
#
# Copyright (c) 1998-2001 Smoot Carl-Mitchell, 2005 Occam's Razor.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Original distribution available at <http://www.occam.com/tools/>.
# 
# $Id: readinfo.pl,v 2.3 2005/11/29 16:57:34 leonvs Exp $
#
# TODO
#

use strict;
use integer;
use FileHandle;				# for autoflush

use Getopt::Std;
use vars qw($opt_h $opt_V $opt_d $opt_i $opt_o $opt_a $opt_F $opt_L $opt_N $opt_P);

# global variables
my $DEBUG = 0;
my $progname;
my $inFile;
my $outFile;
my $OUTPUT_FIELDS_LINE = 0;
my $LINE_NUM = 0;
my $writeOp;
my $PRINT_NULL = 0;
my $PRETTY = 0;
my @extract = ();
my $output;
my @prefix = ();
my @suffix = ();
my @no = ();
my @null = ();
my @fields = ();
my $genName;
my $genStart;
my $genEnd;
my $genIncr;
my $lineNum;

setup();
open OUT, "$writeOp$outFile" or report("ERR", "Can't write to $outFile: $!");
autoflush OUT 1;
processLines($inFile);
if ($PRETTY) {
	prettyPrint($output);
} else {
	print OUT $output;
}
close OUT;

sub setup {
	$progname = $0; $progname =~ s/.*\///;

	my $VERSION = "$progname 2.3\n";

	my $USAGE = <<EOF;
$VERSION
usage:	$progname { -h | -V }
	$progname [-d] [-i filename] [-o filename] [-a] [-F] [-L] [-N] [-P]
		 fieldname [fieldname]...

	-h: Prints this usage statement and exits.
	-V: Prints version number and exits.
	-d: Enables debugging output.
	-i: Specifies the input file (defaults to STDIN).
	-o: Specifies the output file (defaults to STDOUT).
	-a: Appends to output file instead of replacing it.
	-F: Includes the #FIELDS lines in output.
	-L: Prepends line numbers in output.
	-N: Prints "NULL" in empty output fields.
	-P: Makes output prettier by lining up columns.
EOF

	getopts('hVdi:o:aFLNP') or die $USAGE;
	$opt_h and die $USAGE;
	$opt_V and die $VERSION;

	$opt_d and $DEBUG++;

	$inFile = $opt_i || "-";		# default STDIN
	$outFile = $opt_o || ">-";		# default STDOUT
	$writeOp = $opt_a ? ">>" : ">";
	$opt_F and $OUTPUT_FIELDS_LINE++;
	$opt_L and $LINE_NUM++;
	$opt_N and $PRINT_NULL++;
	$opt_P and $PRETTY++;

	@extract=@ARGV;
}

sub processLines {
	my $inFile = shift;
	my $savedLine = "";
	$lineNum = 0;

	# generator definition
	$genStart = 1;
	$genEnd = 1;
	$genIncr = 1;

	open IN, "$inFile" or report("ERR", "Can't read $inFile: $!");
	while (<IN>) {
		chomp;

		if ($savedLine) {
			$_ = "$savedLine$_";
			$savedLine = "";
		}

		# line continues over next line?
		if (/\\$/) {
			$savedLine = $_;
			next;
		}

		# skip blank lines
		next unless /\S/;

		# just do special stuff on non-comments
		next if /^#/ && ! (/^#FIELDS/ ||
				   /^#INLINE/ ||
				   /^#INCLUDE/ ||
				   /^#GENERATOR/ ||
				   /^#END/);

		# change #command to $command
		s/^#/\$/;

		my @in = split /\t+/;
		for ($in[0]) {
			# Extract field definition info
			/^\$FIELDS$/	and do {
						if ($OUTPUT_FIELDS_LINE) {
							newline();
							$output .= "@in\n";
						}
						fieldDef(@in);
						last;
					};

			# $INLINE reads another file and concatenates with input
			/^\$INLINE$/	and do { processInline(@in); last; };

			# $INCLUDE reads another file and processes separately
			/^\$INCLUDE$/	and do { processInclude(@in); last; };

			# $GENERATOR defines a generator
			/^\$GENERATOR$/	and do { defineGenerator(@in); last; };

			# $END terminates a generator
			/^\$END$/	and do { $genName = ""; last; };

			# Process a record
			processRecord(@in);
		}
	}
	close IN;
}

sub newline {
	$lineNum++;
	$LINE_NUM and $output .= "$lineNum\t";
}

sub fieldDef {
	my @line = @_;
	my $nfields = 0;
	my $global = -1;

	# shift off the $FIELDS token
	shift @line;

	foreach my $token (@line) {
		# field is a parameter
		if ($token =~ /\=/) {
			my ($keyword, $value) = split "=", $token;
			if ($keyword eq "prefix") {
				$prefix[$nfields-1] = $value;
			}
			elsif ($keyword eq "suffix") {
				$suffix[$nfields-1] = $value;
			}
			elsif ($keyword eq "no") {
				$no[$nfields-1] = $value;
			}
			elsif ($keyword eq "null") {
				$null[$nfields-1] = $value;
			}
		}
		#field name
		else {
			$fields[$nfields] = $token;
			$prefix[$nfields] = "";
			$suffix[$nfields] = "";
			$no[$nfields] = "";
			$null[$nfields] = "";
			
			# check for GLOBAL special field
			$fields[$nfields] eq "GLOBAL" and $global = $nfields;
			$nfields++;
		}
	}

	# copy in GLOBAL parameters if set
	if ($global != -1) {
		for (my $i=0; $i<$nfields; $i++) {
			$prefix[$i] ||= $prefix[$global];
			$suffix[$i] ||= $suffix[$global];
			$no[$i] ||= $no[$global];
			$null[$i] ||= $null[$global];
		}

		# now shift out the GLOBAL pseudo field
		for (my $i=$global+1; $i<$nfields; $i++) {
			$fields[$i-1] = $fields[$i];
			$prefix[$i-1] = $prefix[$i];
			$suffix[$i-1] = $suffix[$i];
			$no[$i-1] = $no[$i];
			$null[$i-1] = $null[$i];
		}
		$nfields--;
	}

	# quote all special characters in a "no" value
	# this allows any characters in the string
	for (my $i=0; $i<$nfields; $i++) {
		$no[$i] =~ s/(\W)/\\$1/g;
	}

	# check to see all the extract list of fields are in the fields list
	foreach my $field (@extract) {
		my $match = 0;
		for (my $i=0; $i<$nfields; $i++) {
			if ($fields[$i] eq $field) {
				$match = 1;
				last;
			}
		}
		$match or report("ERR", "Field $field not found.");
	}
}

sub processInline {
	my @line = @_;
	my $filename = $line[1];

	# if the filename starts with a | then execute the command
	# with any arguments following and process output
	if ($filename =~ s/^\|//) {
		my $args = join " ", @line[2..$#line];
		processLines("$filename $args |");
	}
	# just continue processing with lines from the inlined file
	else {
		processLines($filename);
	}
}

sub processInclude {
	my @line = @_;
	my $filename = $line[1];

	my $readinfoCmd = "$0 ";
	$readinfoCmd .= "-F " if $OUTPUT_FIELDS_LINE;
	$readinfoCmd .= join " ", @extract;
	report("DEBUG", "readinfo command is $readinfoCmd.");

	# if the filename starts with a | then execute the command
	# with any arguments following and pipe into readinfo
	if ($filename =~ s/^\|//) {
		my $args = join " ", @line[2..$#line];
		open INCLUDE, "$filename $args | $readinfoCmd |" or
			report("ERR", "@line failed: $!");
	}
	# just execute readinfo with the same argument list
	else {
		open INCLUDE, "$readinfoCmd < $filename |" or
			report("ERR", "@line failed: $!");
	}

	while (my $include = <INCLUDE>) {
		newline();
		$output .= $include;
	}

	close INCLUDE;
}

sub defineGenerator {
	my @line = @_;
	$genName = $line[1];
	($genStart, $genEnd, $genIncr) = split /,/, $line[2];

	$genName =~ /^\w+$/ or return 0;
	$genStart =~ /^\d+$/ or return 0;
	$genEnd =~ /^\d+$/ or return 0;
	$genIncr =~ /^\d+$/ or return 0;

	return 1;
}

sub processRecord {
	my @line = @_;

	# see if there is a generator in the line
	# there can only be one
	my $start = 1;
	my $end = 1;
	my $incr = 1;
	if ($genName) {
		$start = $genStart;
		$end = $genEnd;
		$incr = $genIncr;
	}

	for (my $i=$start; $i<=$end; $i += $incr) {
		my $FIRST = 1;
		foreach my $field (@extract) {
			my $j;
			for ($j=0; $j<=$#fields; $j++) {
				last if $fields[$j] eq $field;
			}

			my $null = $null[$j];
			my $no = $no[$j];
			my $prefix = $prefix[$j];
			my $suffix = $suffix[$j];
			my $line = $line[$j] || "";

			# check for null value
			if ($line eq "" or $line eq $null) {
				$prefix = "";
				$suffix = "";
				$line = $PRINT_NULL ? "NULL" : "";
			# if override character exists, remove override
			# character from field and don't use appropriate affix
			} elsif ($no) {
				if ($line =~ /^$no/) {
					$line = substr($line, 1);
					$prefix = "";
				}
				if ($line =~ /$no$/) {
					chop($line);
					$suffix = "";
				}
			}

			# add the generator value
			$genName and $line =~ s/\{$genName\}/$i/;

			if ($FIRST) {
				newline();
				$FIRST = 0;
			}
			else {
				$output .= "\t";
			}
			$output .= "$prefix$line$suffix";
		}
		$output .= "\n";
	}
}

sub prettyPrint {
	my @output = split /\n/, $output;
	my @col = ();
	my @numTabs = ();
	my $ncols = @extract;
	$LINE_NUM and $ncols++;			# extra column for line numbers

	# Get fields into lists per column.
	foreach my $line (@output) {
		my @line = split /\t/, $line, $ncols;
		for (my $i=0; $i<$ncols; $i++) {
			$line[$i] and $col[$i] .= "$line[$i]\t";
		}
	}

	# Find longest element in each column, calculate number of tabs that
	# will fit it. Exclude the last column.
	for (my $i=0; $i<($ncols-1); $i++) {
		my @col = split /\t/, $col[$i];
		my ($longest) = reverse sort byLength @col;
		$numTabs[$i] = ((length $longest) / 8) + 1;
	}

	# Print each line, calculating tabs for each field. Exclude the last
	# field, where we only want a newline following, not trailing tabs.
	foreach my $line (@output) {
		my @line = split /\t/, $line, $ncols;
		for (my $i=0; $i<($ncols-1); $i++) {
			my $numTabs = 0;
			if ($line[$i]) {
				$numTabs =
					$numTabs[$i] - ((length $line[$i]) / 8);
				print OUT $line[$i];
			} else {
				$numTabs = $numTabs[$i];
			}
			print OUT "\t" x $numTabs;
		}
		$line[$ncols-1] and print OUT $line[$ncols-1];
		print OUT "\n";
	}
}

sub byLength {
	length $a <=> length $b;
}

sub report {
	my ($severity, $message) = @_;

	if	($severity eq "ERR")	{ die "$progname: $message\n"; }
	elsif	($severity eq "WARN")	{ warn "$progname: $message\n"; }
	elsif	($severity eq "INFO")	{ print "$progname: $message\n"; }
	elsif	($severity eq "DEBUG")	{ warn "$progname: DEBUG: $message\n"
						if $DEBUG; }
	else	{ die "$progname: Undefined error. $message\n"; }
}
